#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <gdbm.h>
#include <signal.h>
#include "cdict5.h"

#define isBig5CodeHi(c)	((c) >= 0x81 && (c) <=0xFE)

char *word_class[]={"", "", "", "ݴ", "", 
"", "", "ﶯ", "ﶯ", "ϵ", "Ӵ", 
"̾", "", "", "β", ""};

char *freq[]={"", "[γ]", "[]", "[ֻ]", ""};

char *indent[]={"", "  ", "   ", "     ", "  ", "   "};
int  indentnum[]={0, 2, 3, 5, 2, 3};
FILE *pager;
GDBM_FILE indexFD;
char *match[MAX_MATCHES];
int  query_mode, dictFD, L2indexFD;
extern unsigned char *beMarked;

#define	BtoG_bad1 0xa1
#define	BtoG_bad2 0xf5
#define	GtoB_bad1 0xa1
#define	GtoB_bad2 0xbc

extern unsigned char GtoB[], BtoG[];
extern int GtoB_count, BtoG_count;

#define	c1	(unsigned char)(s[0])
#define	c2	(unsigned char)(s[1])

static void g2b(s)
     register char *s;
{
  register unsigned int i;

	if ((c2 >= 0xa1) && (c2 <= 0xfe)) {
		if ((c1 >= 0xa1) && (c1 <= 0xa9)) {
			i = ((c1 - 0xa1) * 94 + (c2 - 0xa1)) * 2;
			s[0] = GtoB[i++];  s[1] = GtoB[i];
			return;
		} else if ((c1 >= 0xb0) && (c1 <= 0xf7)) {
			i = ((c1 - 0xb0 + 9) * 94 + (c2 - 0xa1)) * 2;
			s[0] = GtoB[i++];  s[1] = GtoB[i];
			return;
		}
	}
	s[0] = GtoB_bad1;  s[1] = GtoB_bad2;
}

static void b2g(s)
     register char *s;
{
  register int i;

	if ((c1 >= 0xa1) && (c1 <= 0xf9)) {
		if ((c2 >= 0x40) && (c2 <= 0x7e)) {
			i = ((c1 - 0xa1) * 157 + (c2 - 0x40)) * 2;
			s[0] = BtoG[i++];  s[1] = BtoG[i];
			return;
		} else if ((c2 >= 0xa1) && (c2 <= 0xfe)) {
			i = ((c1 - 0xa1) * 157 + (c2 - 0xa1) + 63) * 2;
			s[0] = BtoG[i++];  s[1] = BtoG[i];
			return;
		}
	}
	s[0] = BtoG_bad1;  s[1] = BtoG_bad2;
}

unsigned char *gb2big(buf)
     unsigned char *buf;
{
   int i;
   for(i=0;buf[i]!='\0';i++) if (buf[i]>=0xa1 && buf[i]<=0xf7) g2b(&buf[i++]);
   return buf;
}

unsigned char *big2gb(buf)
     unsigned char *buf;
{
   int i;
   for(i=0;buf[i]!='\0';i++) if (buf[i]>=0xa1 && buf[i]<=0xfe) b2g(&buf[i++]);
   return buf;
}

#undef c1
#undef c2

static void sig_handle(int signo) {

    switch (signo) {
      case SIGPIPE: pager = stdout;
                    return;
      case SIGSEGV: fprintf(stderr, "caught unexpected SIGSEGV\n");
	            exit(-1);
      case SIGBUS:  fprintf(stderr, "caught unexpected SIGBUS\n");
		    exit(-1);
    }
}

#ifndef LITTLE_ENDIAN
long convertEndian(long i) {
  union {
    int i;
    char c[4];
  } u, v;

  u.i = i;
  v.c[0] = u.c[3];
  v.c[1] = u.c[2];
  v.c[2] = u.c[1];
  v.c[3] = u.c[0];

  return v.i;
}                          
#endif

unsigned char *justify(unsigned char *str, int indent, int stringlen) {
  unsigned char *ptr, *begin, *bp;   /* bp == possible breakpoint */
  int len;

  ptr = str;
  len = COLUMN-indent; 
  begin = str;
  
  while (1) {
    if (stringlen <= len) return big2gb(str);

    while (*ptr) {
      if (ptr-begin > len) break;
      if (*ptr == ' ') 
        bp = ptr;
      else if (isBig5CodeHi(*ptr)) {
        bp = ptr;
        ptr++;
      } 
      ptr++;
    }

    begin = bp+1+indent;  
    if (*bp==' ') bcopy(bp+1, begin, strlen(bp)+1);
    else bcopy(bp, begin, strlen(bp)+1);

    memset(bp+1, ' ', indent);
    *bp = '\n';
    stringlen = strlen(begin);
    ptr = begin;
  } 
}

void printMarked(unsigned char *str) {
  unsigned char *ptr, *strBegin, p;
  int beMarkedLen;
  
  if ((ptr = strstr(str, beMarked)) == NULL) {
    fprintf(pager, "%s\n", str);
    return;
  }

  beMarkedLen = strlen(beMarked);
  strBegin = str;

  while (ptr) {
    p = *ptr;
    *ptr = '\0';
    fprintf(pager, "%s" MARKED_COLOR, strBegin);
    *ptr = p;

    strBegin = ptr + beMarkedLen;
    p = *strBegin;
    *strBegin = '\0';
    fprintf(pager, "%s" ANSI_RESET, ptr);
    *strBegin = p;
    ptr = strstr(strBegin, beMarked);  
  }
  fprintf(pager, "%s\n", strBegin);

}

unsigned int printLine(unsigned char *begin) {
  unsigned char buf[1024], rawchar;
  unsigned int  i, len, offst=2, type;

  len = begin[1];

  if (len==255) { 
    len+=begin[2];
    offst=3;
  }

  type = *begin/0x10;
  if (type==2)
    return len+offst;

  if (type==1) {
    fprintf(pager, KEYWORD_COLOR);
    for (i=offst; i<len+offst; i++) {
      rawchar = begin[i]^(unsigned char)0xa5;
      
      if (rawchar >= 225)      fprintf(pager, ".%c", rawchar-225+'a'); 
      else if (rawchar >= 193) fprintf(pager, ".%c", rawchar-193+'A');
      else if (rawchar==172)   fprintf(pager, "., "); 
      else if (rawchar==160)   fprintf(pager, ". "); 
      else if (rawchar==168)   fprintf(pager, "("); 
      else if (rawchar >=129)  fprintf(pager, ".%c", rawchar-129+'a');
      else fprintf(pager, "%c", rawchar);
    } 
    fprintf(pager, ANSI_RESET " %s\n", freq[*begin-0x10]);

  }
  else {
    for (i=0; i<len; i++) {
      buf[i] = begin[i+offst]^(char)0xa5;
      if (buf[i]=='#') {
        buf[i]=0;
        break;
      }
    }        
    buf[len]=0;

    if (*begin == 0xC4 || *begin == 0xD8) 
      fprintf(pager, INFLEXION_COLOR "(%s)" ANSI_RESET "\n" 
                   , justify(buf, 0, len));

    else if (*begin == 0x84 || type==0x5) 
      fprintf(pager, IDIOM_COLOR "%s" ANSI_RESET " \n"
		   , justify(buf, 0, len));

    else if (type == 0xD || type == 0xA || type == 0xE || type == 0x9) {
      int i=0, j;
      switch (type) {
        case 0xD: i = *begin-0xDC; break;
        case 0xA: i = *begin-0xA5; break;
        case 0x9: i = *begin-0x9D; break;
	case 0xE: i = *begin-0xE4; break;  
      }

      if (i<0 || i>=sizeof(indentnum)/sizeof(indentnum[0])) i=0;
      fprintf(pager, "%s" EXAMPLE_COLOR "%s" ANSI_RESET "\n"
		   , indent[i], justify(buf, indentnum[i], len));

    }

    else if (type == 0xC) {
      int j;

      if (*begin>=0xC8) {
        j = *begin - 0xC8;
        fprintf(pager, "%s" OTHER_INFO_COLOR "%s" ANSI_RESET "\n"
                     , indent[j], justify(buf, indentnum[j], len));
      }
      else {
        j = *begin - 0xC0;
        fprintf(pager, indent[j]);
        if (query_mode & QUERY_CHINESE) 
          printMarked(justify(buf, indentnum[j], len));
        else
          fprintf(pager, "%s\n", justify(buf, indentnum[j], len));
      }
    }   
    else if (type == 0x8){
        int j;
      
        j = *begin - 0x81;
        if (j<0 || j >= sizeof(indentnum)/sizeof(indentnum[0])) j=0;

        fprintf(pager, indent[j]);
        if (query_mode & QUERY_CHINESE) 
          printMarked(justify(buf, indentnum[j], len));
        else
          fprintf(pager, "%s\n", justify(buf, indentnum[j], len));
    }
    else
      fprintf(pager, "%s\n", justify(buf, 0, len));
  }
  
  return len+offst;
}

int processArgs(unsigned char *buf, unsigned char *word) {
  int i, j, mode;

  mode = 0;

  if (buf[strlen(buf)-1] == '*' ) mode|=WILDCARD_MODE;

  if (isBig5CodeHi(buf[0])) mode|=QUERY_CHINESE;

  if (!(mode & QUERY_CHINESE)) {
    j=0;
    i=0;
    while (buf[i]) {
      if (isalpha(buf[i])) word[j++] = tolower(buf[i]);
      i++;
    }           
    word[j] = 0;
  }
  else 
    strcpy(word, strtok(buf, "*"));

  return mode;
}

int findAllMatches(char *word) {
  char *prefix, *tmp, *tmp2;
  char indexfile[] = "?.idx";
  datum key, data;  
  int   totalMatch = 0;

  prefix = (unsigned char*) malloc(strlen(word)+2);
  sprintf(prefix, "\n%s", word);

  *indexfile = *word;
  key.dptr = indexfile;
  key.dsize = sizeof(indexfile);

  data = gdbm_fetch(indexFD, key);
  if (data.dptr == NULL) {
    fprintf(stderr, "index file corrupted: entry \"%s\" not found\n", indexfile);
    exit(-1);
  }                      

  tmp2 = data.dptr;

  while (totalMatch < MAX_MATCHES) {
    tmp = strstr(tmp2, prefix);
    if (tmp) {
      tmp2 = strchr(tmp+1, '\n');
      *tmp2 = 0;
      match[totalMatch++] = strdup(tmp+1);
      *tmp2 = '\n';
    }
    else {
      if (totalMatch == 0) {
        fprintf(stderr, "no word begins with \"%s\"\n", word); 
        exit(-1);
      }
      break;
    }
  }

  free(data.dptr);
  free(prefix);     

  return totalMatch;
}

void printResult(unsigned char *buf, int buflen) {
  int index = 0;
  
  fprintf(pager, "\n");
  do {
    if (buf[index]/0x10 == 6) {   
      fprintf(pager, WORD_CLASS_COLOR "<<%s>>" ANSI_RESET "\n"
		   , word_class[buf[index]-0x60]);
      index++;
    }
    else if (buf[index] == 0x50) {
      printf(WORD_CLASS_COLOR "<<ͬ>>" ANSI_RESET "\n");
      index++;
    }
    else index += (printLine(buf+index));
  } while (buf[index]/0x10 != 1 && index <= buflen-1); 
}

void printAllResult(int totalMatch) {
  int i, j, data_len, L2index, L3index, index;
  unsigned char *tmp, *buf;
  datum key, data;

  pager = stdout;
  for (j = 0; j < totalMatch; j++) {
    key.dptr  = match[j];
    key.dsize = strlen(match[j])+1;

    if(!strcmp(key.dptr,"headman")||!strcmp(key.dptr,"inweave")) continue;

    data = gdbm_fetch(indexFD, key);
    if (data.dptr == NULL) {
      if (query_mode & WILDCARD_MODE) 
        fprintf(stderr, "index file corrupted: entry \"%s\" not found\n", match[j]);
      else  
        fprintf(stderr, "\"%s\" not found\n", match[j]);  
      break;
    } 

    if (j == 0) {      
#ifdef PAGER
      if ((pager = popen(PAGER, "w")) == NULL) {
        fprintf(stderr, "cannot use %s as pager\n\n", PAGER);
        pager = stdout;
      }
#else
      pager = stdout;
#endif

      if (query_mode & WILDCARD_MODE)
        fprintf(pager, "%dַϲѯ:\n", totalMatch); 
    }

    tmp = data.dptr;
    for (i = 0; i < (data.dsize)/2; i++, tmp+=2) {

      L2index = (tmp[0] + tmp[1]*0x100)*4;
      
      lseek(L2indexFD, L2index, 0);
      if (read(L2indexFD, &L3index, 4) < 0) {  
        fprintf(pager, "dictionary file corrupted\n");
        exit(-1);
      }
#ifdef LITTLE_ENDIAN
      L3index &= 0x1FFFFFL;
#else
      L3index = convertEndian(L3index) & 0x1FFFFFL;
#endif      

      if (read(L2indexFD, &data_len, 4) < 0)
        data_len = L3index + 4096;
      else
#ifdef LITTLE_ENDIAN
      data_len &= 0x1FFFFFL;
#else
      data_len = convertEndian(data_len) & 0x1FFFFFL;
#endif
                       
      lseek(dictFD, L3index, 0);
      buf = (unsigned char *)malloc(data_len - L3index + 10);
      data_len = read(dictFD, buf, data_len - L3index + 10);

      if (*buf/0x10 != 1 || data_len <= 0) {
        fprintf(stderr, "dictionary file corrupted: entry \"%s\"\n", match[j]);
        break;
      }

      printResult(buf, data_len);
      free(buf);
    }

    free(data.dptr);
  }                      

  fprintf(pager, "\n");
  if (pager != stdout)  pclose(pager); 
}

void queryEnglish(char *word) {
  int totalMatch;
  unsigned char *index_path, *dict_path, *L2index_path;

  index_path = (unsigned char*)malloc(sizeof(INDEX_PATH)+10);
  sprintf(index_path, INDEX_PATH "/Index.gdbm");

  indexFD = gdbm_open(index_path, 0, GDBM_READER, 0600, NULL);
  if (indexFD == NULL) {
    fprintf(stderr, "cannot access index file %s\n", index_path);
    gdbm_strerror(gdbm_errno);
    exit(-1);
  }
  free(index_path);          

  if (query_mode & WILDCARD_MODE)
    totalMatch = findAllMatches(word);
  else {
    totalMatch = 1;
    match[0] = word;
  }

  dict_path = (unsigned char*)malloc(sizeof(DICT_PATH)+10);
  L2index_path = (unsigned char*)malloc(sizeof(DICT_PATH)+10); 

  sprintf(dict_path, DICT_PATH "/%c.d50", *word);
  sprintf(L2index_path, DICT_PATH "/%c.i50", *word);

  dictFD = open(dict_path, O_RDONLY);
  L2indexFD = open(L2index_path, O_RDONLY);

  if (L2indexFD < 0 || dictFD < 0) {
    fprintf(pager, "cannot access dictionary file %s or %s\n",
                    L2index_path, dict_path);
    exit(-1);
  }

  free(dict_path);
  free(L2index_path);
  printAllResult(totalMatch);   
  close(dictFD);
  close(L2indexFD);
}
  
void main(int argc, char **argv) {
  int i;
  unsigned char buf[4096], word[4096];

  if (argc < 2) {
    fprintf(stderr, "\nUsage: %s word[*]\n\n find all words matching \"word\"\n"
		  , *argv);
    return;
  }

  buf[0]=0;
  for (i = 1; i < argc; i++) strcat(buf, argv[i]);
 
  query_mode = processArgs(gb2big(buf), word);
    
  if ((query_mode & WILDCARD_MODE) && word[1] == '\0') {
    fprintf(stderr, "need longer prefix for wildcard mode\n");
    exit(-1);
  }

  signal(SIGPIPE, sig_handle);
  signal(SIGSEGV, sig_handle);
  signal(SIGBUS, sig_handle);

  if (query_mode & QUERY_CHINESE) 
    queryChinese(word);
  else
    queryEnglish(word);
}
